# (c) Copyright IBM Corp. 2025

import os
import time
from typing import Generator

import pytest
from confluent_kafka import (
    Consumer,
    KafkaException,
    Producer,
)
from confluent_kafka.admin import AdminClient, NewTopic
from mock import patch, Mock
from opentelemetry.trace import SpanKind
from opentelemetry.trace.span import format_span_id

from instana.configurator import config
from instana.options import StandardOptions
from instana.singletons import agent, tracer
from instana.util.config import parse_ignored_endpoints_from_yaml
from tests.helpers import get_first_span_by_filter, testenv
from instana.instrumentation.kafka import confluent_kafka_python
from instana.instrumentation.kafka.confluent_kafka_python import (
    clear_context,
    save_consumer_span_into_context,
    close_consumer_span,
    trace_kafka_close,
    consumer_span,
)
from instana.span.span import InstanaSpan


class TestConfluentKafka:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # Kafka admin client
        self.kafka_config = {"bootstrap.servers": testenv["kafka_bootstrap_servers"][0]}
        self.kafka_client = AdminClient(self.kafka_config)

        try:
            _ = self.kafka_client.create_topics(  # noqa: F841
                [
                    NewTopic(
                        testenv["kafka_topic"],
                        num_partitions=1,
                        replication_factor=1,
                    ),
                    NewTopic(
                        testenv["kafka_topic"] + "_1",
                        num_partitions=1,
                        replication_factor=1,
                    ),
                    NewTopic(
                        testenv["kafka_topic"] + "_2",
                        num_partitions=1,
                        replication_factor=1,
                    ),
                    NewTopic(
                        testenv["kafka_topic"] + "_3",
                        num_partitions=1,
                        replication_factor=1,
                    ),
                ]
            )
        except KafkaException:
            pass

        # Kafka producer
        self.producer = Producer(self.kafka_config)
        agent.options = StandardOptions()
        yield
        # teardown
        # Clear spans before resetting options
        self.recorder.clear_spans()

        # Clear context
        clear_context()

        # Close connections
        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"],
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
                testenv["kafka_topic"] + "_3",
            ]
        )
        time.sleep(3)

    def test_trace_confluent_kafka_produce(self) -> None:
        with tracer.start_as_current_span("test"):
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes")
            self.producer.flush(timeout=10)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.CLIENT
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "produce"

    def test_trace_confluent_kafka_consume(self) -> None:
        agent.options.set_trace_configurations()
        # Produce some events
        self.producer.produce(testenv["kafka_topic"], value=b"raw_bytes1")
        self.producer.flush(timeout=30)

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        with tracer.start_as_current_span("test"):
            msgs = consumer.consume(num_messages=1, timeout=60)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

    def test_trace_confluent_kafka_poll(self) -> None:
        # Produce some events
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        with tracer.start_as_current_span("test"):
            msg = consumer.poll(timeout=3)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        def filter(span):
            return span.n == "kafka" and span.data["kafka"]["access"] == "poll"

        kafka_span = get_first_span_by_filter(spans, filter)

        def filter(span):
            return span.n == "sdk" and span.data["sdk"]["name"] == "test"

        test_span = get_first_span_by_filter(spans, filter)

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "poll"

    def test_trace_confluent_kafka_error(self) -> None:
        # Consume the events
        consumer_config = {"bootstrap.servers": ["some_inexistent_host:9094"]}
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe(["inexistent_kafka_topic"])

        with tracer.start_as_current_span("test"):
            consumer.consume(-10)

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[len(spans) - 1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert kafka_span.ec == 1

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert not kafka_span.data["kafka"]["service"]
        assert kafka_span.data["kafka"]["access"] == "consume"
        assert (
            kafka_span.data["kafka"]["error"]
            == "num_messages must be between 0 and 1000000 (1M)"
        )

    @patch.dict(os.environ, {"INSTANA_IGNORE_ENDPOINTS": "kafka"})
    def test_ignore_confluent_kafka(self) -> None:
        agent.options.set_trace_configurations()
        with tracer.start_as_current_span("test"):
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes")
            self.producer.flush(timeout=10)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

    @patch.dict(os.environ, {"INSTANA_IGNORE_ENDPOINTS": "kafka:produce"})
    def test_ignore_confluent_kafka_producer(self) -> None:
        agent.options.set_trace_configurations()
        with tracer.start_as_current_span("test-span"):
            # Produce some events
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes2")
            self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])
        consumer.consume(num_messages=2, timeout=60)

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

    @patch.dict(os.environ, {"INSTANA_IGNORE_ENDPOINTS": "kafka:consume"})
    def test_ignore_confluent_kafka_consumer(self) -> None:
        agent.options.set_trace_configurations()
        # Produce some events
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        with tracer.start_as_current_span("test-span"):
            # Consume the events
            consumer_config = self.kafka_config.copy()
            consumer_config["group.id"] = "my-group"
            consumer_config["auto.offset.reset"] = "earliest"

            consumer = Consumer(consumer_config)
            consumer.subscribe([testenv["kafka_topic"]])
            consumer.consume(num_messages=2, timeout=60)

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

    @patch.dict(
        os.environ,
        {
            "INSTANA_IGNORE_ENDPOINTS_PATH": "tests/util/test_configuration-1.yaml",
        },
    )
    def test_ignore_confluent_specific_topic(self) -> None:
        agent.options.set_trace_configurations()
        self.kafka_client.create_topics(  # noqa: F841
            [
                NewTopic(
                    testenv["kafka_topic"] + "_1",
                    num_partitions=1,
                    replication_factor=1,
                ),
            ]
        )

        with tracer.start_as_current_span("test-span"):
            # Produce some events
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
            self.producer.produce(testenv["kafka_topic"] + "_1", b"raw_bytes1")
            self.producer.flush()

            # Consume the events
            consumer_config = self.kafka_config.copy()
            consumer_config["group.id"] = "my-group"
            consumer_config["auto.offset.reset"] = "earliest"

            consumer = Consumer(consumer_config)
            consumer.subscribe([testenv["kafka_topic"], testenv["kafka_topic"] + "_1"])
            consumer.consume(num_messages=2, timeout=60)

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 3

        span_to_be_filtered = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["service"] == "span-topic",
        )
        assert span_to_be_filtered not in filtered_spans

        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"] + "_1",
            ]
        )

    def test_ignore_confluent_specific_topic_with_config_file(self) -> None:
        agent.options.ignore_endpoints = parse_ignored_endpoints_from_yaml(
            "tests/util/test_configuration-1.yaml"
        )

        with tracer.start_as_current_span("test-span"):
            # Produce some events
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
            self.producer.flush()

            # Consume the events
            consumer_config = self.kafka_config.copy()
            consumer_config["group.id"] = "my-group"
            consumer_config["auto.offset.reset"] = "earliest"

            consumer = Consumer(consumer_config)
            consumer.subscribe([testenv["kafka_topic"]])
            consumer.consume(num_messages=1, timeout=60)
        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

    def test_confluent_kafka_consumer_root_exit(self) -> None:
        agent.options.allow_exit_as_root = True

        self.producer.produce(testenv["kafka_topic"] + "_1", b"raw_bytes")
        self.producer.produce(testenv["kafka_topic"] + "_2", b"raw_bytes")
        self.producer.flush(timeout=10)

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
            ]
        )

        consumer.consume(num_messages=2, timeout=60)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        consumer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "consume"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        consumer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "consume"
            and span.data["kafka"]["service"] == "span-topic_2",
        )

        # same trace id, different span ids
        assert producer_span_1.t == consumer_span_1.t
        assert producer_span_1.s == consumer_span_1.p
        assert producer_span_1.s != consumer_span_1.s

        assert producer_span_2.t == consumer_span_2.t
        assert producer_span_2.s == consumer_span_2.p
        assert producer_span_2.s != consumer_span_2.s

        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
            ]
        )

    def test_confluent_kafka_poll_root_exit_with_trace_correlation(self) -> None:
        agent.options.allow_exit_as_root = True
        agent.options.set_trace_configurations()

        # Produce some events
        self.producer.produce(testenv["kafka_topic"] + "-poll", b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"] + "-poll"])

        msg = consumer.poll(timeout=30)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        producer_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic-poll",
        )

        poll_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic-poll",
        )

        # Same traceId
        assert producer_span.t == poll_span.t
        assert producer_span.s == poll_span.p
        assert producer_span.s != poll_span.s

    def test_confluent_kafka_poll_root_exit_without_trace_correlation(self) -> None:
        agent.options.allow_exit_as_root = True
        agent.options.kafka_trace_correlation = False

        # Produce some events
        self.producer.produce(f'{testenv["kafka_topic"]}-wo-tc', b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([f'{testenv["kafka_topic"]}-wo-tc'])

        msg = consumer.poll(timeout=30)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        producer_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == f'{testenv["kafka_topic"]}-wo-tc',
        )

        poll_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == f'{testenv["kafka_topic"]}-wo-tc',
        )

        # Different traceId
        assert producer_span.t != poll_span.t
        assert producer_span.s != poll_span.p
        assert producer_span.s != poll_span.s

    def test_confluent_kafka_poll_root_exit_error(self) -> None:
        agent.options.allow_exit_as_root = True
        agent.options.set_trace_configurations()

        # Produce some events
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        msg = consumer.poll(timeout="wrong_value")  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        poll_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka" and span.data["kafka"]["access"] == "poll",
        )
        assert poll_span.data["kafka"]["error"] == "must be real number, not str"

    @patch.dict(os.environ, {"INSTANA_ALLOW_ROOT_EXIT_SPAN": "1"})
    def test_confluent_kafka_downstream_suppression(self) -> None:
        config["tracing"]["ignore_endpoints"] = {
            "kafka": [
                {"methods": ["produce"], "endpoints": [f"{testenv['kafka_topic']}_1"]},
                {
                    "methods": ["consume"],
                    "endpoints": [f"{testenv['kafka_topic']}_2"],
                },
            ]
        }
        agent.options.set_trace_configurations()

        self.kafka_client.create_topics(  # noqa: F841
            [
                NewTopic(
                    testenv["kafka_topic"] + "_1",
                    num_partitions=1,
                    replication_factor=1,
                ),
                NewTopic(
                    testenv["kafka_topic"] + "_2",
                    num_partitions=1,
                    replication_factor=1,
                ),
            ]
        )

        self.producer.produce(testenv["kafka_topic"] + "_1", b"raw_bytes1")
        self.producer.produce(testenv["kafka_topic"] + "_2", b"raw_bytes2")
        self.producer.flush(timeout=10)

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
            ]
        )

        messages = consumer.consume(num_messages=2, timeout=60)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        consumer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "consume"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        consumer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "consume"
            and span.data["kafka"]["service"] == "span-topic_2",
        )

        assert producer_span_1
        # consumer has been suppressed
        assert not consumer_span_1
        assert not consumer_span_2

        for message in messages:
            if message.topic() == "span-topic_1":
                assert message.headers() == [("x_instana_l_s", b"0")]
            else:
                assert message.headers() == [
                    ("x_instana_l_s", b"1"),
                    ("x_instana_t", format_span_id(producer_span_2.t).encode("utf-8")),
                    ("x_instana_s", format_span_id(producer_span_2.s).encode("utf-8")),
                ]

        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
            ]
        )

    def test_save_consumer_span_into_context(self, span: "InstanaSpan") -> None:
        """Test save_consumer_span_into_context function."""
        # Verify initial state
        assert consumer_span.get(None) is None
        assert confluent_kafka_python.consumer_token is None

        # Save span into context
        save_consumer_span_into_context(span)

        # Verify token is stored
        assert confluent_kafka_python.consumer_token is not None

    def test_close_consumer_span_recording_span(self, span: "InstanaSpan") -> None:
        """Test close_consumer_span with a recording span."""
        # Save span into context first
        save_consumer_span_into_context(span)
        assert confluent_kafka_python.consumer_token is not None

        # Verify span is recording
        assert span.is_recording()

        # Close the span
        close_consumer_span(span)

        # Verify span was ended and context cleared
        assert not span.is_recording()
        assert consumer_span.get(None) is None
        assert confluent_kafka_python.consumer_token is None

    def test_clear_context(self, span: "InstanaSpan") -> None:
        """Test clear_context function."""
        # Save span into context
        save_consumer_span_into_context(span)

        # Verify context has data
        assert consumer_span.get(None) == span
        assert confluent_kafka_python.consumer_token is not None

        # Clear context
        clear_context()

        # Verify all context is cleared
        assert consumer_span.get(None) is None
        assert confluent_kafka_python.consumer_token is None

    def test_trace_kafka_close_exception_handling(self, span: "InstanaSpan") -> None:
        """Test trace_kafka_close handles exceptions and still cleans up spans."""
        # Save span into context
        save_consumer_span_into_context(span)

        # Verify span is in context
        assert consumer_span.get(None) == span
        assert confluent_kafka_python.consumer_token is not None

        # Mock a wrapped function that raises an exception
        mock_wrapped = Mock(side_effect=Exception("Close operation failed"))
        mock_instance = Mock()

        # Call trace_kafka_close - it should handle the exception gracefully
        # and still clean up the span
        trace_kafka_close(mock_wrapped, mock_instance, (), {})

        # Verify the wrapped function was called
        mock_wrapped.assert_called_once_with()

        # Verify that despite the exception, the span was cleaned up
        assert consumer_span.get(None) is None
        assert confluent_kafka_python.consumer_token is None

        # Verify span was ended
        assert not span.is_recording()



# (c) Copyright IBM Corp. 2025

from typing import List, Optional
import pytest

from instana.util.span_utils import get_operation_specifiers


@pytest.mark.parametrize(
    "span_name, expected_result",
    [
        ("something", ["", ""]),
        ("redis", ["command", ""]),
        ("dynamodb", ["op", ""]),
        ("kafka", ["access", "service"]),
    ],
)
def test_get_operation_specifiers(
    span_name: str,
    expected_result: Optional[List[str]],
) -> None:
    operation_specifier, service_specifier = get_operation_specifiers(span_name)
    assert operation_specifier == expected_result[0]
    assert service_specifier == expected_result[1]



# (c) Copyright IBM Corp. 2025

import os
from typing import Generator

import pytest
from kafka import KafkaConsumer, KafkaProducer
from kafka.admin import KafkaAdminClient, NewTopic
from kafka.errors import TopicAlreadyExistsError
from mock import patch
from opentelemetry.trace import SpanKind
from opentelemetry.trace.span import format_span_id

from instana.configurator import config
from instana.options import StandardOptions
from instana.singletons import agent, tracer
from instana.util.config import parse_ignored_endpoints_from_yaml
from tests.helpers import get_first_span_by_filter, testenv

from instana.instrumentation.kafka import kafka_python
from instana.instrumentation.kafka.kafka_python import (
    clear_context,
    save_consumer_span_into_context,
    close_consumer_span,
    consumer_span,
)
from instana.span.span import InstanaSpan


class TestKafkaPython:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # Kafka admin client
        self.kafka_client = KafkaAdminClient(
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            client_id="test_kafka_python",
        )

        try:
            self.kafka_client.create_topics(
                [
                    NewTopic(
                        name=testenv["kafka_topic"],
                        num_partitions=1,
                        replication_factor=1,
                    ),
                    NewTopic(
                        name=testenv["kafka_topic"] + "_1",
                        num_partitions=1,
                        replication_factor=1,
                    ),
                    NewTopic(
                        name=testenv["kafka_topic"] + "_2",
                        num_partitions=1,
                        replication_factor=1,
                    ),
                    NewTopic(
                        name=testenv["kafka_topic"] + "_3",
                        num_partitions=1,
                        replication_factor=1,
                    ),
                ]
            )
        except TopicAlreadyExistsError:
            pass

        # Kafka producer
        self.producer = KafkaProducer(
            bootstrap_servers=testenv["kafka_bootstrap_servers"]
        )
        agent.options = StandardOptions()
        yield
        # teardown
        # Ensure that allow_exit_as_root has the default value"""
        agent.options.allow_exit_as_root = False
        # Close connections
        self.producer.close()

        # Clear context
        clear_context()

        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"],
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
                testenv["kafka_topic"] + "_3",
            ]
        )
        self.kafka_client.close()

    def test_trace_kafka_python_send(self) -> None:
        with tracer.start_as_current_span("test"):
            future = self.producer.send(testenv["kafka_topic"], b"raw_bytes")

        _ = future.get(timeout=10)  # noqa: F841

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.CLIENT
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "send"

    def test_trace_kafka_python_consume(self) -> None:
        # Produce some events
        self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        with tracer.start_as_current_span("test"):
            for msg in consumer:
                if msg is None:
                    break

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        def filter(span):
            return span.n == "kafka" and span.data["kafka"]["access"] == "consume"

        kafka_span = get_first_span_by_filter(spans, filter)

        def filter(span):
            return span.n == "sdk" and span.data["sdk"]["name"] == "test"

        test_span = get_first_span_by_filter(spans, filter)

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "consume"

    def test_trace_kafka_python_poll(self) -> None:
        # Produce some events
        self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        with tracer.start_as_current_span("test"):
            msg = consumer.poll(timeout_ms=3000)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        def filter(span):
            return span.n == "kafka" and span.data["kafka"]["access"] == "poll"

        kafka_span = get_first_span_by_filter(spans, filter)

        def filter(span):
            return span.n == "sdk" and span.data["sdk"]["name"] == "test"

        test_span = get_first_span_by_filter(spans, filter)

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "poll"

    def test_trace_kafka_python_error(self) -> None:
        consumer = KafkaConsumer(
            "inexistent_kafka_topic",
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",
            enable_auto_commit=False,
            consumer_timeout_ms=1000,
        )

        with tracer.start_as_current_span("test"):
            consumer._client = None

            try:
                for msg in consumer:
                    if msg is None:
                        break
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        def filter(span):
            return span.n == "kafka" and span.data["kafka"]["access"] == "consume"

        kafka_span = get_first_span_by_filter(spans, filter)

        def filter(span):
            return span.n == "sdk" and span.data["sdk"]["name"] == "test"

        test_span = get_first_span_by_filter(spans, filter)

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert kafka_span.ec == 1

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == "inexistent_kafka_topic"
        assert kafka_span.data["kafka"]["access"] == "consume"
        assert (
            kafka_span.data["kafka"]["error"]
            == "'NoneType' object has no attribute 'poll'"
        )

    def consume_from_topic(self, topic_name: str) -> None:
        consumer = KafkaConsumer(
            topic_name,
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",
            enable_auto_commit=False,
            consumer_timeout_ms=1000,
        )
        with tracer.start_as_current_span("test"):
            for msg in consumer:
                if msg is None:
                    break

        consumer.close()

    @patch.dict(os.environ, {"INSTANA_IGNORE_ENDPOINTS": "kafka"})
    def test_ignore_kafka(self) -> None:
        agent.options.set_trace_configurations()
        with tracer.start_as_current_span("test"):
            self.producer.send(testenv["kafka_topic"], b"raw_bytes")
            self.producer.flush()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

    @patch.dict(os.environ, {"INSTANA_IGNORE_ENDPOINTS": "kafka:send"})
    def test_ignore_kafka_producer(self) -> None:
        agent.options.set_trace_configurations()
        with tracer.start_as_current_span("test-span"):
            # Produce some events
            self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
            self.producer.send(testenv["kafka_topic"], b"raw_bytes2")
            self.producer.flush()

        # Consume the events manually
        # consume_from_topic not used due to to not create sdk span
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",
            enable_auto_commit=False,
            consumer_timeout_ms=1000,
        )
        for msg in consumer:
            if msg is None:
                break

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

    @patch.dict(os.environ, {"INSTANA_IGNORE_ENDPOINTS": "kafka:consume"})
    def test_ignore_kafka_consumer(self) -> None:
        agent.options.set_trace_configurations()
        # Produce some events
        self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        # Consume the events
        self.consume_from_topic(testenv["kafka_topic"])

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

    @patch.dict(
        os.environ,
        {
            "INSTANA_IGNORE_ENDPOINTS_PATH": "tests/util/test_configuration-1.yaml",
        },
    )
    def test_ignore_specific_topic(self) -> None:
        agent.options.set_trace_configurations()
        with tracer.start_as_current_span("test-span"):
            # Produce some events
            self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
            self.producer.send(testenv["kafka_topic"] + "_1", b"raw_bytes1")
            self.producer.flush()

            # Consume the events
            self.consume_from_topic(testenv["kafka_topic"])
            self.consume_from_topic(testenv["kafka_topic"] + "_1")

        spans = self.recorder.queued_spans()
        assert len(spans) == 7

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 6

        span_to_be_filtered = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["service"] == "span-topic",
        )
        assert span_to_be_filtered not in filtered_spans

    def test_ignore_specific_topic_with_config_file(self) -> None:
        agent.options.ignore_endpoints = parse_ignored_endpoints_from_yaml(
            "tests/util/test_configuration-1.yaml"
        )

        # Produce some events
        self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        self.consume_from_topic(testenv["kafka_topic"])

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

    def test_kafka_consumer_root_exit(self) -> None:
        agent.options.allow_exit_as_root = True

        self.producer.send(testenv["kafka_topic"], b"raw_bytes")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        for msg in consumer:
            if msg is None:
                break

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        producer_span = spans[0]
        consumer_span = spans[1]

        assert producer_span.s
        assert producer_span.n == "kafka"
        assert producer_span.data["kafka"]["access"] == "send"
        assert producer_span.data["kafka"]["service"] == "span-topic"

        assert consumer_span.s
        assert consumer_span.n == "kafka"
        assert consumer_span.data["kafka"]["access"] == "consume"
        assert consumer_span.data["kafka"]["service"] == "span-topic"

        assert producer_span.t == consumer_span.t

    def test_kafka_poll_root_exit_with_trace_correlation(self) -> None:
        agent.options.allow_exit_as_root = True

        self.producer.send(testenv["kafka_topic"] + "_1", b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"] + "_2", b"raw_bytes2")
        self.producer.send(testenv["kafka_topic"] + "_3", b"raw_bytes3")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )
        topics = [
            testenv["kafka_topic"] + "_1",
            testenv["kafka_topic"] + "_2",
            testenv["kafka_topic"] + "_3",
        ]
        consumer.subscribe(topics)

        messages = consumer.poll(timeout_ms=1000)  # noqa: F841
        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 6

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        producer_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        poll_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        poll_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        poll_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        assert producer_span_1.n == "kafka"
        assert producer_span_1.data["kafka"]["access"] == "send"
        assert producer_span_1.data["kafka"]["service"] == "span-topic_1"

        assert producer_span_2.n == "kafka"
        assert producer_span_2.data["kafka"]["access"] == "send"
        assert producer_span_2.data["kafka"]["service"] == "span-topic_2"

        assert producer_span_3.n == "kafka"
        assert producer_span_3.data["kafka"]["access"] == "send"
        assert producer_span_3.data["kafka"]["service"] == "span-topic_3"

        assert poll_span_1.n == "kafka"
        assert poll_span_1.data["kafka"]["access"] == "poll"
        assert poll_span_1.data["kafka"]["service"] == "span-topic_1"

        assert poll_span_2.n == "kafka"
        assert poll_span_2.data["kafka"]["access"] == "poll"
        assert poll_span_2.data["kafka"]["service"] == "span-topic_2"

        assert poll_span_3.n == "kafka"
        assert poll_span_3.data["kafka"]["access"] == "poll"
        assert poll_span_3.data["kafka"]["service"] == "span-topic_3"

        # same trace id, different span ids
        assert producer_span_1.t == poll_span_1.t
        assert producer_span_1.s != poll_span_1.s

        assert producer_span_2.t == poll_span_2.t
        assert producer_span_2.s != poll_span_2.s

        assert producer_span_3.t == poll_span_3.t
        assert producer_span_3.s != poll_span_3.s

    def test_kafka_poll_root_exit_without_trace_correlation(self) -> None:
        agent.options.allow_exit_as_root = True
        agent.options.kafka_trace_correlation = False

        self.producer.send(testenv["kafka_topic"] + "_1", b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"] + "_2", b"raw_bytes2")
        self.producer.send(testenv["kafka_topic"] + "_3", b"raw_bytes3")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )
        topics = [
            testenv["kafka_topic"] + "_1",
            testenv["kafka_topic"] + "_2",
            testenv["kafka_topic"] + "_3",
        ]
        consumer.subscribe(topics)

        messages = consumer.poll(timeout_ms=1000)  # noqa: F841
        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 6

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        producer_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        poll_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        poll_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        poll_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        assert producer_span_1.n == "kafka"
        assert producer_span_1.data["kafka"]["access"] == "send"
        assert producer_span_1.data["kafka"]["service"] == "span-topic_1"

        assert producer_span_2.n == "kafka"
        assert producer_span_2.data["kafka"]["access"] == "send"
        assert producer_span_2.data["kafka"]["service"] == "span-topic_2"

        assert producer_span_3.n == "kafka"
        assert producer_span_3.data["kafka"]["access"] == "send"
        assert producer_span_3.data["kafka"]["service"] == "span-topic_3"

        assert poll_span_1.n == "kafka"
        assert poll_span_1.data["kafka"]["access"] == "poll"
        assert poll_span_1.data["kafka"]["service"] == "span-topic_1"

        assert poll_span_2.n == "kafka"
        assert poll_span_2.data["kafka"]["access"] == "poll"
        assert poll_span_2.data["kafka"]["service"] == "span-topic_2"

        assert poll_span_3.n == "kafka"
        assert poll_span_3.data["kafka"]["access"] == "poll"
        assert poll_span_3.data["kafka"]["service"] == "span-topic_3"

        # different trace id and span ids
        assert producer_span_1.t != poll_span_1.t
        assert producer_span_1.s != poll_span_1.s

        assert producer_span_2.t != poll_span_2.t
        assert producer_span_2.s != poll_span_2.s

        assert producer_span_3.t != poll_span_3.t
        assert producer_span_3.s != poll_span_3.s

        for topic_partition, partition_messages in messages.items():
            for message in partition_messages:
                assert not message.headers

    @patch.dict(os.environ, {"INSTANA_ALLOW_ROOT_EXIT_SPAN": "1"})
    def test_kafka_downstream_suppression(self) -> None:
        config["tracing"]["ignore_endpoints"] = {
            "kafka": [
                {"methods": ["send"], "endpoints": [f"{testenv['kafka_topic']}_1"]},
                {
                    "methods": ["consume"],
                    "endpoints": [f"{testenv['kafka_topic']}_2"],
                },
            ]
        }
        agent.options.set_trace_configurations()

        self.producer.send(testenv["kafka_topic"] + "_1", b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"] + "_2", b"raw_bytes2")
        self.producer.send(testenv["kafka_topic"] + "_3", b"raw_bytes3")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )
        topics = [
            testenv["kafka_topic"] + "_1",
            testenv["kafka_topic"] + "_2",
            testenv["kafka_topic"] + "_3",
        ]
        consumer.subscribe(topics)

        messages = consumer.poll(timeout_ms=1000)  # noqa: F841
        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 5

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        producer_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        poll_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        poll_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        assert producer_span_1.n == "kafka"
        assert producer_span_1.data["kafka"]["access"] == "send"
        assert producer_span_1.data["kafka"]["service"] == "span-topic_1"

        assert producer_span_2.n == "kafka"
        assert producer_span_2.data["kafka"]["access"] == "send"
        assert producer_span_2.data["kafka"]["service"] == "span-topic_2"

        assert producer_span_3.n == "kafka"
        assert producer_span_3.data["kafka"]["access"] == "send"
        assert producer_span_3.data["kafka"]["service"] == "span-topic_3"

        assert poll_span_2.n == "kafka"
        assert poll_span_2.data["kafka"]["access"] == "poll"
        assert poll_span_2.data["kafka"]["service"] == "span-topic_2"

        assert poll_span_3.n == "kafka"
        assert poll_span_3.data["kafka"]["access"] == "poll"
        assert poll_span_3.data["kafka"]["service"] == "span-topic_3"

        # same trace id, different span ids
        assert producer_span_2.t == poll_span_2.t
        assert producer_span_2.s != poll_span_2.s

        assert producer_span_3.t == poll_span_3.t
        assert producer_span_3.s != poll_span_3.s

        for topic_partition, partition_messages in messages.items():
            for message in partition_messages:
                if message.topic == "span-topic_1":
                    assert message.headers == [("x_instana_l_s", b"0")]
                elif message.topic == "span-topic_2":
                    assert message.headers == [
                        ("x_instana_l_s", b"1"),
                        (
                            "x_instana_t",
                            format_span_id(producer_span_2.t).encode("utf-8"),
                        ),
                        (
                            "x_instana_s",
                            format_span_id(producer_span_2.s).encode("utf-8"),
                        ),
                    ]

    def test_save_consumer_span_into_context(self, span: "InstanaSpan") -> None:
        """Test save_consumer_span_into_context function."""
        # Verify initial state
        assert consumer_span.get(None) is None
        assert kafka_python.consumer_token is None

        # Save span into context
        save_consumer_span_into_context(span)

        # Verify span is saved in context variable
        assert consumer_span.get(None) == span
        # Verify token is stored
        assert kafka_python.consumer_token is not None

    def test_close_consumer_span_recording_span(self, span: "InstanaSpan") -> None:
        """Test close_consumer_span with a recording span."""
        # Save span into context first
        save_consumer_span_into_context(span)
        assert kafka_python.consumer_token is not None

        # Verify span is recording
        assert span.is_recording()

        # Close the span
        close_consumer_span(span)

        # Verify span was ended and context cleared
        assert not span.is_recording()
        assert consumer_span.get(None) is None
        assert kafka_python.consumer_token is None

    def test_clear_context(self, span: "InstanaSpan") -> None:
        """Test clear_context function."""
        # Save span into context
        save_consumer_span_into_context(span)

        # Verify context has data
        assert consumer_span.get(None) == span
        assert kafka_python.consumer_token is not None

        # Clear context
        clear_context()

        # Verify all context is cleared
        assert consumer_span.get(None) is None
        assert kafka_python.consumer_token is None



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import boto3
import pytest
import urllib3
from typing import Generator

from moto import mock_aws

import tests.apps.flask_app  # noqa: F401
from instana.singletons import tracer, agent
from tests.helpers import get_first_span_by_filter, get_first_span_by_name, testenv

pwd = os.path.dirname(os.path.abspath(__file__))


class TestSqs:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Setup and Teardown"""
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.mock = mock_aws()
        self.mock.start()
        self.sqs = boto3.client("sqs", region_name="us-east-1")
        self.http_client = urllib3.PoolManager()
        yield
        # Stop Moto after each test
        self.mock.stop()
        agent.options.allow_exit_as_root = False

    def test_vanilla_create_queue(self) -> None:
        result = self.sqs.create_queue(
            QueueName="SQS_QUEUE_NAME",
            Attributes={"DelaySeconds": "60", "MessageRetentionPeriod": "86400"},
        )
        assert result["ResponseMetadata"]["HTTPStatusCode"] == 200

    def test_send_message(self) -> None:
        # Create the Queue:
        response = self.sqs.create_queue(
            QueueName="SQS_QUEUE_NAME",
            Attributes={"DelaySeconds": "60", "MessageRetentionPeriod": "600"},
        )

        assert response["QueueUrl"]
        queue_url = response["QueueUrl"]

        with tracer.start_as_current_span("test"):
            response = self.sqs.send_message(
                QueueUrl=queue_url,
                DelaySeconds=10,
                MessageAttributes={
                    "Website": {
                        "DataType": "String",
                        "StringValue": "https://www.instana.com",
                    },
                },
                MessageBody=(
                    "Monitor any application, service, or request "
                    "with Instana Application Performance Monitoring"
                ),
            )

        assert response["MessageId"]

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span

        boto_span = get_first_span_by_name(spans, "boto3")
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec

        assert boto_span.data["boto3"]["op"] == "SendMessage"
        assert boto_span.data["boto3"]["ep"] == "https://sqs.us-east-1.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"

        payload = {
            "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/SQS_QUEUE_NAME",
            "DelaySeconds": 10,
            "MessageAttributes": {
                "Website": {
                    "DataType": "String",
                    "StringValue": "https://www.instana.com",
                }
            },
            "MessageBody": "Monitor any application, service, or request with Instana Application Performance Monitoring",
        }
        assert boto_span.data["boto3"]["payload"] == payload

        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://sqs.us-east-1.amazonaws.com:443/SendMessage"
        )

    def test_send_message_as_root_exit_span(self) -> None:
        # Create the Queue:
        response = self.sqs.create_queue(
            QueueName="SQS_QUEUE_NAME",
            Attributes={"DelaySeconds": "60", "MessageRetentionPeriod": "600"},
        )

        assert response["QueueUrl"]
        agent.options.allow_exit_as_root = True
        queue_url = response["QueueUrl"]

        response = self.sqs.send_message(
            QueueUrl=queue_url,
            DelaySeconds=10,
            MessageAttributes={
                "Website": {
                    "DataType": "String",
                    "StringValue": "https://www.instana.com",
                },
            },
            MessageBody=(
                "Monitor any application, service, or request "
                "with Instana Application Performance Monitoring"
            ),
        )

        assert response["MessageId"]

        spans = self.recorder.queued_spans()
        assert len(spans) == 1
        boto_span = spans[0]
        assert boto_span
        assert boto_span.n == "boto3"
        assert not boto_span.p
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "SendMessage"
        assert boto_span.data["boto3"]["ep"] == "https://sqs.us-east-1.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"

        payload = {
            "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/SQS_QUEUE_NAME",
            "DelaySeconds": 10,
            "MessageAttributes": {
                "Website": {
                    "DataType": "String",
                    "StringValue": "https://www.instana.com",
                }
            },
            "MessageBody": "Monitor any application, service, or request with Instana Application Performance Monitoring",
        }
        assert boto_span.data["boto3"]["payload"] == payload

        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://sqs.us-east-1.amazonaws.com:443/SendMessage"
        )

    def test_app_boto3_sqs(self) -> None:
        with tracer.start_as_current_span("test"):
            self.http_client.request("GET", testenv["flask_server"] + "/boto3/sqs")

        spans = self.recorder.queued_spans()
        assert len(spans) == 5

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span

        http_span = get_first_span_by_name(spans, "urllib3")
        assert http_span

        wsgi_span = get_first_span_by_name(spans, "wsgi")
        assert wsgi_span

        bcq_span = get_first_span_by_filter(
            spans,
            (
                lambda span: span.n == "boto3"
                and span.data["boto3"]["op"] == "CreateQueue"
            ),
        )
        assert bcq_span

        bsm_span = get_first_span_by_filter(
            spans,
            (
                lambda span: span.n == "boto3"
                and span.data["boto3"]["op"] == "SendMessage"
            ),
        )
        assert bsm_span

        assert http_span.t == test_span.t
        assert http_span.p == test_span.s

        assert wsgi_span.t == test_span.t
        assert wsgi_span.p == http_span.s

        assert bcq_span.t == test_span.t
        assert bcq_span.p == wsgi_span.s

        assert bsm_span.t == test_span.t
        assert bsm_span.p == wsgi_span.s

    def test_request_header_capture_before_call(self) -> None:
        # Create the Queue:
        response = self.sqs.create_queue(
            QueueName="SQS_QUEUE_NAME",
            Attributes={"DelaySeconds": "60", "MessageRetentionPeriod": "600"},
        )

        assert response["QueueUrl"]

        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        # Access the event system on the S3 client
        event_system = self.sqs.meta.events

        request_headers = {"X-Capture-This": "this", "X-Capture-That": "that"}

        # Create a function that adds custom headers
        def add_custom_header_before_call(params, **kwargs):
            params["headers"].update(request_headers)

        # Register the function to before-call event.
        event_system.register(
            "before-call.sqs.SendMessage", add_custom_header_before_call
        )

        queue_url = response["QueueUrl"]
        with tracer.start_as_current_span("test"):
            response = self.sqs.send_message(
                QueueUrl=queue_url,
                DelaySeconds=10,
                MessageAttributes={
                    "Website": {
                        "DataType": "String",
                        "StringValue": "https://www.instana.com",
                    },
                },
                MessageBody=(
                    "Monitor any application, service, or request "
                    "with Instana Application Performance Monitoring"
                ),
            )

        assert response["MessageId"]

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span

        boto_span = get_first_span_by_name(spans, "boto3")
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec

        assert boto_span.data["boto3"]["op"] == "SendMessage"
        assert boto_span.data["boto3"]["ep"] == "https://sqs.us-east-1.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"

        payload = {
            "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/SQS_QUEUE_NAME",
            "DelaySeconds": 10,
            "MessageAttributes": {
                "Website": {
                    "DataType": "String",
                    "StringValue": "https://www.instana.com",
                }
            },
            "MessageBody": "Monitor any application, service, or request with Instana Application Performance Monitoring",
        }
        assert boto_span.data["boto3"]["payload"] == payload

        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://sqs.us-east-1.amazonaws.com:443/SendMessage"
        )

        assert "X-Capture-This" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-That"] == "that"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture_before_sign(self) -> None:
        # Create the Queue:
        response = self.sqs.create_queue(
            QueueName="SQS_QUEUE_NAME",
            Attributes={"DelaySeconds": "60", "MessageRetentionPeriod": "600"},
        )

        assert response["QueueUrl"]

        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Custom-1", "X-Custom-2"]

        # Access the event system on the S3 client
        event_system = self.sqs.meta.events

        request_headers = {"X-Custom-1": "Value1", "X-Custom-2": "Value2"}

        # Create a function that adds custom headers
        def add_custom_header_before_sign(request, **kwargs):
            for name, value in request_headers.items():
                request.headers.add_header(name, value)

        # Register the function to before-sign event.
        event_system.register_first(
            "before-sign.sqs.SendMessage", add_custom_header_before_sign
        )

        queue_url = response["QueueUrl"]
        with tracer.start_as_current_span("test"):
            response = self.sqs.send_message(
                QueueUrl=queue_url,
                DelaySeconds=10,
                MessageAttributes={
                    "Website": {
                        "DataType": "String",
                        "StringValue": "https://www.instana.com",
                    },
                },
                MessageBody=(
                    "Monitor any application, service, or request "
                    "with Instana Application Performance Monitoring"
                ),
            )

        assert response["MessageId"]

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span

        boto_span = get_first_span_by_name(spans, "boto3")
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec

        assert boto_span.data["boto3"]["op"] == "SendMessage"
        assert boto_span.data["boto3"]["ep"] == "https://sqs.us-east-1.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"

        payload = {
            "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/SQS_QUEUE_NAME",
            "DelaySeconds": 10,
            "MessageAttributes": {
                "Website": {
                    "DataType": "String",
                    "StringValue": "https://www.instana.com",
                }
            },
            "MessageBody": "Monitor any application, service, or request with Instana Application Performance Monitoring",
        }
        assert boto_span.data["boto3"]["payload"] == payload

        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://sqs.us-east-1.amazonaws.com:443/SendMessage"
        )

        assert "X-Custom-1" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Custom-1"] == "Value1"
        assert "X-Custom-2" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Custom-2"] == "Value2"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_response_header_capture(self) -> None:
        # Create the Queue:
        response = self.sqs.create_queue(
            QueueName="SQS_QUEUE_NAME",
            Attributes={"DelaySeconds": "60", "MessageRetentionPeriod": "600"},
        )

        assert response["QueueUrl"]

        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        # Access the event system on the S3 client
        event_system = self.sqs.meta.events

        response_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        # Create a function that sets the custom headers in the after-call event.
        def modify_after_call_args(parsed, **kwargs):
            parsed["ResponseMetadata"]["HTTPHeaders"].update(response_headers)

        # Register the function to an event
        event_system.register("after-call.sqs.SendMessage", modify_after_call_args)

        queue_url = response["QueueUrl"]
        with tracer.start_as_current_span("test"):
            response = self.sqs.send_message(
                QueueUrl=queue_url,
                DelaySeconds=10,
                MessageAttributes={
                    "Website": {
                        "DataType": "String",
                        "StringValue": "https://www.instana.com",
                    },
                },
                MessageBody=(
                    "Monitor any application, service, or request "
                    "with Instana Application Performance Monitoring"
                ),
            )

        assert response["MessageId"]

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span

        boto_span = get_first_span_by_name(spans, "boto3")
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec

        assert boto_span.data["boto3"]["op"] == "SendMessage"
        assert boto_span.data["boto3"]["ep"] == "https://sqs.us-east-1.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"

        payload = {
            "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/SQS_QUEUE_NAME",
            "DelaySeconds": 10,
            "MessageAttributes": {
                "Website": {
                    "DataType": "String",
                    "StringValue": "https://www.instana.com",
                }
            },
            "MessageBody": "Monitor any application, service, or request with Instana Application Performance Monitoring",
        }
        assert boto_span.data["boto3"]["payload"] == payload

        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://sqs.us-east-1.amazonaws.com:443/SendMessage"
        )

        assert "X-Capture-This-Too" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2025

from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Discovery:
    pid: int = 0  # the PID of this process
    name: Optional[str] = None  # the name of the executable
    args: Optional[List[str]] = None  # the command line arguments
    fd: int = -1  # the file descriptor of the socket associated with the connection to the agent for this HTTP request
    inode: str = ""  # the inode of the socket associated with the connection to the agent for this HTTP request



# (c) Copyright IBM Corp. 2025

from typing import Tuple


def get_operation_specifiers(span_name: str) -> Tuple[str, str]:
    """Get the specific operation specifier for the given span."""
    operation_specifier_key = ""
    service_specifier_key = ""
    if span_name == "redis":
        operation_specifier_key = "command"
    elif span_name == "dynamodb":
        operation_specifier_key = "op"
    elif span_name == "kafka":
        operation_specifier_key = "access"
        service_specifier_key = "service"
    return operation_specifier_key, service_specifier_key



# (c) Copyright IBM Corp. 2025

import logging
import os
from typing import TYPE_CHECKING, Generator

import pytest
from yaml import YAMLError

from instana.util.config import (
    get_disable_trace_configurations_from_yaml,
    parse_ignored_endpoints_from_yaml,
)
from instana.util.config_reader import ConfigReader

if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture


class TestConfigReader:
    @pytest.fixture(autouse=True)
    def _resource(
        self,
        caplog: "LogCaptureFixture",
    ) -> Generator[None, None, None]:
        yield
        caplog.clear()
        if "INSTANA_CONFIG_PATH" in os.environ:
            os.environ.pop("INSTANA_CONFIG_PATH")

    def test_config_reader_null(self, caplog: "LogCaptureFixture") -> None:
        config_reader = ConfigReader(os.environ.get("INSTANA_CONFIG_PATH", ""))
        assert config_reader.file_path == ""
        assert config_reader.data == {}
        assert "ConfigReader: No configuration file specified" in caplog.messages

    def test_config_reader_default(self) -> None:
        filename = "tests/util/test_configuration-1.yaml"
        os.environ["INSTANA_CONFIG_PATH"] = filename
        config_reader = ConfigReader(os.environ.get("INSTANA_CONFIG_PATH", ""))
        assert config_reader.file_path == filename
        assert "tracing" in config_reader.data
        assert len(config_reader.data["tracing"]) == 2

    def test_config_reader_file_not_found_error(
        self, caplog: "LogCaptureFixture"
    ) -> None:
        filename = "tests/util/test_configuration-3.yaml"
        os.environ["INSTANA_CONFIG_PATH"] = filename
        config_reader = ConfigReader(os.environ.get("INSTANA_CONFIG_PATH", ""))
        assert config_reader.file_path == filename
        assert config_reader.data == {}
        assert (
            f"ConfigReader: Configuration file has not found: {filename}"
            in caplog.messages
        )

    def test_config_reader_yaml_error(
        self, caplog: "LogCaptureFixture", mocker: "MockerFixture"
    ) -> None:
        filename = "tests/util/test_configuration-1.yaml"
        exception_message = "BLAH"
        mocker.patch(
            "instana.util.config_reader.yaml.safe_load",
            side_effect=YAMLError(exception_message),
        )

        config_reader = ConfigReader(filename)  # noqa: F841
        assert (
            f"ConfigReader: Error parsing YAML file: {exception_message}"
            in caplog.messages
        )

    def test_load_configuration_with_tracing(self, caplog: "LogCaptureFixture") -> None:
        caplog.set_level(logging.DEBUG, logger="instana")

        ignore_endpoints = parse_ignored_endpoints_from_yaml(
            "tests/util/test_configuration-1.yaml"
        )
        # test with tracing
        assert ignore_endpoints == [
            "redis.get",
            "redis.type",
            "dynamodb.query",
            "kafka.consume.span-topic",
            "kafka.consume.topic1",
            "kafka.consume.topic2",
            "kafka.send.span-topic",
            "kafka.send.topic1",
            "kafka.send.topic2",
            "kafka.consume.topic3",
            "kafka.*.span-topic",
            "kafka.*.topic4",
        ]

        os.environ["INSTANA_CONFIG_PATH"] = "tests/util/test_configuration-1.yaml"
        disabled_spans, enabled_spans = get_disable_trace_configurations_from_yaml()
        # Check disabled_spans list
        assert "logging" in disabled_spans
        assert "databases" in disabled_spans
        assert "redis" not in disabled_spans
        assert "redis" in enabled_spans

        assert (
            'Please use "tracing" instead of "com.instana.tracing" for local configuration file.'
            not in caplog.messages
        )

    def test_load_configuration_legacy(self, caplog: "LogCaptureFixture") -> None:
        caplog.set_level(logging.DEBUG, logger="instana")

        ignore_endpoints = parse_ignored_endpoints_from_yaml(
            "tests/util/test_configuration-2.yaml"
        )
        assert ignore_endpoints == [
            "redis.get",
            "redis.type",
            "dynamodb.query",
            "kafka.send.*",
            "kafka.consume.span-topic",
            "kafka.consume.topic1",
            "kafka.consume.topic2",
            "kafka.send.span-topic",
            "kafka.send.topic1",
            "kafka.send.topic2",
            "kafka.consume.topic3",
            "kafka.*.span-topic",
            "kafka.*.topic4",
        ]

        os.environ["INSTANA_CONFIG_PATH"] = "tests/util/test_configuration-2.yaml"
        disabled_spans, enabled_spans = get_disable_trace_configurations_from_yaml()
        # Check disabled_spans list
        assert "logging" in disabled_spans
        assert "databases" in disabled_spans
        assert "redis" not in disabled_spans
        assert "redis" in enabled_spans

        assert (
            'Please use "tracing" instead of "com.instana.tracing" for local configuration file.'
            in caplog.messages
        )



import asyncio
from typing import Any, Generator

import aioamqp
import pytest

from instana.singletons import tracer
from tests.helpers import testenv
from aioamqp.properties import Properties
from aioamqp.envelope import Envelope

testenv["rabbitmq_host"] = "127.0.0.1"
testenv["rabbitmq_port"] = 5672


class TestAioamqp:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        yield
        self.loop.run_until_complete(self.delete_queue())
        if self.loop.is_running():
            self.loop.close()

    async def delete_queue(self) -> None:
        _, protocol = await aioamqp.connect(
            testenv["rabbitmq_host"],
            testenv["rabbitmq_port"],
        )
        channel = await protocol.channel()
        await channel.queue_delete("message_queue")
        await asyncio.sleep(1)

    async def publish_message(self) -> None:
        transport, protocol = await aioamqp.connect(
            testenv["rabbitmq_host"],
            testenv["rabbitmq_port"],
        )
        channel = await protocol.channel()

        await channel.queue_declare(queue_name="message_queue")

        message = "Instana test message"
        await channel.basic_publish(
            message.encode(), exchange_name="", routing_key="message_queue"
        )

        await protocol.close()
        transport.close()

    async def consume_message(self) -> None:
        async def callback(
            channel: Any,
            body: bytes,
            envelope: Envelope,
            properties: Properties,
        ) -> None:
            with tracer.start_as_current_span("callback-span"):
                await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)

        _, protocol = await aioamqp.connect(
            testenv["rabbitmq_host"], testenv["rabbitmq_port"]
        )
        channel = await protocol.channel()
        await channel.queue_declare(queue_name="message_queue")
        await channel.basic_consume(callback, queue_name="message_queue", no_ack=False)

    def test_basic_publish(self) -> None:
        with tracer.start_as_current_span("test-span"):
            self.loop.run_until_complete(self.publish_message())

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        publisher_span = spans[0]
        test_span = spans[1]

        assert publisher_span.n == "amqp"
        assert publisher_span.data["amqp"]["command"] == "publish"
        assert publisher_span.data["amqp"]["routingkey"] == "message_queue"
        assert publisher_span.data["amqp"]["connection"] == "127.0.0.1:5672"

        assert publisher_span.p == test_span.s

        assert test_span.n == "sdk"
        assert not test_span.p

    def test_basic_consumer(self) -> None:
        with tracer.start_as_current_span("test-span"):
            self.loop.run_until_complete(self.publish_message())
            self.loop.run_until_complete(self.consume_message())

        spans = self.recorder.queued_spans()

        assert len(spans) == 4

        publisher_span = spans[0]
        callback_span = spans[1]
        consumer_span = spans[2]
        test_span = spans[3]

        assert publisher_span.n == "amqp"
        assert publisher_span.data["amqp"]["command"] == "publish"
        assert publisher_span.data["amqp"]["routingkey"] == "message_queue"
        assert publisher_span.data["amqp"]["connection"] == "127.0.0.1:5672"
        assert publisher_span.p == test_span.s

        assert callback_span.n == "sdk"
        assert callback_span.data["sdk"]["name"] == "callback-span"
        assert callback_span.data["sdk"]["type"] == "intermediate"
        assert callback_span.p == consumer_span.s

        assert consumer_span.n == "amqp"
        assert consumer_span.data["amqp"]["command"] == "consume"
        assert consumer_span.data["amqp"]["routingkey"] == "message_queue"
        assert consumer_span.data["amqp"]["connection"] == "127.0.0.1:5672"
        assert (
            consumer_span.data["amqp"]["connection"]
            == publisher_span.data["amqp"]["connection"]
        )
        assert consumer_span.p == test_span.s

        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test-span"



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016

"""
Option classes for the in-process Instana agent

The description and hierarchy of the classes in this file are as follows:

BaseOptions - base class for all environments.  Holds settings common to all.
  - StandardOptions - The options class used when running directly on a host/node with an Instana agent
  - ServerlessOptions - Base class for serverless environments.  Holds settings common to all serverless environments.
    - AWSLambdaOptions - Options class for AWS Lambda.  Holds settings specific to AWS Lambda.
    - AWSFargateOptions - Options class for AWS Fargate.  Holds settings specific to AWS Fargate.
    - GCROptions - Options class for Google cloud Run.  Holds settings specific to GCR.
"""

import logging
import os
from typing import Any, Dict, Sequence

from instana.configurator import config
from instana.log import logger
from instana.util.config import (
    SPAN_TYPE_TO_CATEGORY,
    get_disable_trace_configurations_from_env,
    get_disable_trace_configurations_from_local,
    get_disable_trace_configurations_from_yaml,
    is_truthy,
    parse_ignored_endpoints,
    parse_ignored_endpoints_from_yaml,
    parse_span_disabling,
)
from instana.util.runtime import determine_service_name


class BaseOptions(object):
    """Base class for all option classes.  Holds items common to all"""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        self.debug = False
        self.log_level = logging.WARN
        self.service_name = determine_service_name()
        self.extra_http_headers = None
        self.allow_exit_as_root = False
        self.ignore_endpoints = []
        self.kafka_trace_correlation = True

        # disabled_spans lists all categories and types that should be disabled
        self.disabled_spans = []
        # enabled_spans lists all categories and types that should be enabled, preceding disabled_spans
        self.enabled_spans = []

        self.set_trace_configurations()

        # Defaults
        self.secrets_matcher = "contains-ignore-case"
        self.secrets_list = ["key", "pass", "secret"]

        # Env var format: <matcher>:<secret>[,<secret>]
        self.secrets = os.environ.get("INSTANA_SECRETS", None)

        if self.secrets is not None:
            parts = self.secrets.split(":")
            if len(parts) == 2:
                self.secrets_matcher = parts[0]
                self.secrets_list = parts[1].split(",")
            else:
                logger.warning(
                    f"Couldn't parse INSTANA_SECRETS env var: {self.secrets}"
                )

        self.__dict__.update(kwds)

    def set_trace_configurations(self) -> None:
        """
        Set tracing configurations from the environment variables and config file.
        @return: None
        """
        # Use self.configurations to not read local configuration file
        # in set_tracing method
        if "INSTANA_DEBUG" in os.environ:
            self.log_level = logging.DEBUG
            self.debug = True

        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            self.extra_http_headers = (
                str(os.environ["INSTANA_EXTRA_HTTP_HEADERS"]).lower().split(";")
            )

        # Check if either of the environment variables is truthy
        if is_truthy(os.environ.get("INSTANA_ALLOW_EXIT_AS_ROOT", None)) or is_truthy(
            os.environ.get("INSTANA_ALLOW_ROOT_EXIT_SPAN", None)
        ):
            self.allow_exit_as_root = True

        # The priority is as follows:
        # environment variables > in-code configuration >
        # > agent config (configuration.yaml) > default value
        if "INSTANA_IGNORE_ENDPOINTS" in os.environ:
            self.ignore_endpoints = parse_ignored_endpoints(
                os.environ["INSTANA_IGNORE_ENDPOINTS"]
            )
        elif "INSTANA_IGNORE_ENDPOINTS_PATH" in os.environ:
            self.ignore_endpoints = parse_ignored_endpoints_from_yaml(
                os.environ["INSTANA_IGNORE_ENDPOINTS_PATH"]
            )
        elif (
            isinstance(config.get("tracing"), dict)
            and "ignore_endpoints" in config["tracing"]
        ):
            self.ignore_endpoints = parse_ignored_endpoints(
                config["tracing"]["ignore_endpoints"],
            )

        if "INSTANA_KAFKA_TRACE_CORRELATION" in os.environ:
            self.kafka_trace_correlation = is_truthy(
                os.environ["INSTANA_KAFKA_TRACE_CORRELATION"]
            )
        elif isinstance(config.get("tracing"), dict) and "kafka" in config["tracing"]:
            self.kafka_trace_correlation = config["tracing"]["kafka"].get(
                "trace_correlation", True
            )

        self.set_disable_trace_configurations()

    def set_disable_trace_configurations(self) -> None:
        disabled_spans = []
        enabled_spans = []

        # The precedence is as follows:
        # environment variables > in-code (local) config > agent config (configuration.yaml)
        # For the env vars: INSTANA_TRACING_DISABLE > INSTANA_CONFIG_PATH
        if "INSTANA_TRACING_DISABLE" in os.environ:
            disabled_spans, enabled_spans = get_disable_trace_configurations_from_env()
        elif "INSTANA_CONFIG_PATH" in os.environ:
            disabled_spans, enabled_spans = get_disable_trace_configurations_from_yaml()
        else:
            # In-code (local) config
            # The agent config (configuration.yaml) is handled in StandardOptions.set_disable_tracing()
            disabled_spans, enabled_spans = (
                get_disable_trace_configurations_from_local()
            )

        self.disabled_spans.extend(disabled_spans)
        self.enabled_spans.extend(enabled_spans)

    def is_span_disabled(self, category=None, span_type=None) -> bool:
        """
        Check if a span is disabled based on its category and type.

        Args:
            category (str): The span category (e.g., "logging", "databases")
            span_type (str): The span type (e.g., "redis", "kafka")

        Returns:
            bool: True if the span is disabled, False otherwise
        """
        # If span_type is provided, check if it's disabled
        if span_type and span_type in self.disabled_spans:
            return True

        # If category is provided directly, check if it's disabled
        if category and category in self.disabled_spans:
            return True

        # If span_type is provided but not explicitly configured,
        # check if its parent category is disabled. Also check for the precedence rules
        if span_type and span_type in SPAN_TYPE_TO_CATEGORY:
            parent_category = SPAN_TYPE_TO_CATEGORY[span_type]
            if (
                parent_category in self.disabled_spans
                and span_type not in self.enabled_spans
            ):
                return True

        # Default: not disabled
        return False


class StandardOptions(BaseOptions):
    """The options class used when running directly on a host/node with an Instana agent"""

    AGENT_DEFAULT_HOST = "localhost"
    AGENT_DEFAULT_PORT = 42699

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(StandardOptions, self).__init__()

        self.agent_host = os.environ.get("INSTANA_AGENT_HOST", self.AGENT_DEFAULT_HOST)
        self.agent_port = os.environ.get("INSTANA_AGENT_PORT", self.AGENT_DEFAULT_PORT)

        if not isinstance(self.agent_port, int):
            self.agent_port = int(self.agent_port)

    def set_secrets(self, secrets: Dict[str, Any]) -> None:
        """
        Set the secret option from the agent config.
        @param secrets: dictionary of secrets
        @return: None
        """
        self.secrets_matcher = secrets["matcher"]
        self.secrets_list = secrets["list"]

    def set_extra_headers(self, extra_headers: Dict[str, Any]) -> None:
        """
        Set the extra headers option from the agent config, which uses the legacy configuration setting.
        @param extra_headers: dictionary of headers
        @return: None
        """
        if self.extra_http_headers is None:
            self.extra_http_headers = extra_headers
        else:
            self.extra_http_headers.extend(extra_headers)
        logger.info(
            f"Will also capture these custom headers: {self.extra_http_headers}"
        )

    def set_tracing(self, tracing: Dict[str, Any]) -> None:
        """
        Set tracing options from the agent config.
        @param tracing: tracing configuration dictionary
        @return: None
        """
        if "ignore-endpoints" in tracing and not self.ignore_endpoints:
            self.ignore_endpoints = parse_ignored_endpoints(tracing["ignore-endpoints"])

        if "kafka" in tracing:
            if (
                "INSTANA_KAFKA_TRACE_CORRELATION" not in os.environ
                and not (
                    isinstance(config.get("tracing"), dict)
                    and "kafka" in config["tracing"]
                )
                and "trace-correlation" in tracing["kafka"]
            ):
                self.kafka_trace_correlation = is_truthy(
                    tracing["kafka"].get("trace-correlation", True)
                )

            if (
                "header-format" in tracing["kafka"]
                and tracing["kafka"]["header-format"] == "binary"
            ):
                logger.warning(
                    "Binary header format for Kafka is deprecated. Please use string header format."
                )

        if "extra-http-headers" in tracing:
            self.extra_http_headers = tracing["extra-http-headers"]

        # Handle span disabling configuration
        if "disable" in tracing:
            self.set_disable_tracing(tracing["disable"])

    def set_disable_tracing(self, tracing_config: Sequence[Dict[str, Any]]) -> None:
        # The precedence is as follows:
        # environment variables > in-code (local) config > agent config (configuration.yaml)
        if (
            "INSTANA_TRACING_DISABLE" not in os.environ
            and "INSTANA_CONFIG_PATH" not in os.environ
            and not (
                isinstance(config.get("tracing"), dict)
                and "disable" in config["tracing"]
            )
        ):
            # agent config (configuration.yaml)
            disabled_spans, enabled_spans = parse_span_disabling(tracing_config)
            self.disabled_spans.extend(disabled_spans)
            self.enabled_spans.extend(enabled_spans)

    def set_from(self, res_data: Dict[str, Any]) -> None:
        """
        Set the source identifiers given to use by the Instana Host agent.
        @param res_data: source identifiers provided as announce response
        @return: None
        """
        if not res_data or not isinstance(res_data, dict):
            logger.debug(f"options.set_from: Wrong data type - {type(res_data)}")
            return

        if "secrets" in res_data:
            self.set_secrets(res_data["secrets"])

        if "tracing" in res_data:
            self.set_tracing(res_data["tracing"])

        else:
            if "extraHeaders" in res_data:
                self.set_extra_headers(res_data["extraHeaders"])


class ServerlessOptions(BaseOptions):
    """Base class for serverless environments.  Holds settings common to all serverless environments."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(ServerlessOptions, self).__init__()

        self.agent_key = os.environ.get("INSTANA_AGENT_KEY", None)
        self.endpoint_url = os.environ.get("INSTANA_ENDPOINT_URL", None)

        # Remove any trailing slash (if any)
        if self.endpoint_url is not None and self.endpoint_url[-1] == "/":
            self.endpoint_url = self.endpoint_url[:-1]

        if "INSTANA_DISABLE_CA_CHECK" in os.environ:
            self.ssl_verify = False
        else:
            self.ssl_verify = True

        proxy = os.environ.get("INSTANA_ENDPOINT_PROXY", None)
        if proxy is None:
            self.endpoint_proxy = {}
        else:
            self.endpoint_proxy = {"https": proxy}

        timeout_in_ms = os.environ.get("INSTANA_TIMEOUT", None)
        if timeout_in_ms is None:
            self.timeout = 0.8
        else:
            # Convert the value from milliseconds to seconds for the requests package
            try:
                self.timeout = int(timeout_in_ms) / 1000
            except ValueError:
                logger.warning(
                    f"Likely invalid INSTANA_TIMEOUT={timeout_in_ms} value.  Using default."
                )
                logger.warning(
                    "INSTANA_TIMEOUT should specify timeout in milliseconds.  See "
                    "https://www.instana.com/docs/reference/environment_variables/#serverless-monitoring"
                )
                self.timeout = 0.8

        value = os.environ.get("INSTANA_LOG_LEVEL", None)
        if value is not None:
            try:
                value = value.lower()
                if value == "debug":
                    self.log_level = logging.DEBUG
                elif value == "info":
                    self.log_level = logging.INFO
                elif value == "warn" or value == "warning":
                    self.log_level = logging.WARNING
                elif value == "error":
                    self.log_level = logging.ERROR
                else:
                    logger.warning(f"Unknown INSTANA_LOG_LEVEL specified: {value}")
            except Exception:
                logger.debug("BaseAgent.update_log_level: ", exc_info=True)


class AWSLambdaOptions(ServerlessOptions):
    """Options class for AWS Lambda.  Holds settings specific to AWS Lambda."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(AWSLambdaOptions, self).__init__()


class AWSFargateOptions(ServerlessOptions):
    """Options class for AWS Fargate.  Holds settings specific to AWS Fargate."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(AWSFargateOptions, self).__init__()

        self.tags = None
        tag_list = os.environ.get("INSTANA_TAGS", None)
        if tag_list is not None:
            try:
                self.tags = dict()
                tags = tag_list.split(",")
                for tag_and_value in tags:
                    parts = tag_and_value.split("=")
                    length = len(parts)
                    if length == 1:
                        self.tags[parts[0]] = None
                    elif length == 2:
                        self.tags[parts[0]] = parts[1]
            except Exception:
                logger.debug(f"Error parsing INSTANA_TAGS env var: {tag_list}")

        self.zone = os.environ.get("INSTANA_ZONE", None)


class EKSFargateOptions(AWSFargateOptions):
    """Options class for EKS Pods on AWS Fargate. Holds settings specific to EKS Pods on AWS Fargate."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(EKSFargateOptions, self).__init__()


class GCROptions(ServerlessOptions):
    """Options class for Google Cloud Run.  Holds settings specific to Google Cloud Run."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(GCROptions, self).__init__()



# (c) Copyright IBM Corp. 2025

import pytest
from typing import Generator, TYPE_CHECKING
import asyncio
from aio_pika import Message, connect, connect_robust

from instana.singletons import agent, tracer

if TYPE_CHECKING:
    from instana.span.readable_span import ReadableSpan


class TestAioPika:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.queue_name = "test.queue"
        yield
        # teardown
        self.loop.run_until_complete(self.delete_queue())
        if self.loop.is_running():
            self.loop.close()
        # Ensure that allow_exit_as_root has the default value
        agent.options.allow_exit_as_root = False

    async def publish_message(self, params_combination: str = "both_args") -> None:
        # Perform connection
        connection = await connect()

        async with connection:
            # Creating a channel
            channel = await connection.channel()

            # Declaring queue
            queue_name = self.queue_name
            queue = await channel.declare_queue(queue_name)

            # Declaring exchange
            exchange = await channel.declare_exchange("test.exchange")
            await queue.bind(exchange, routing_key=queue_name)

            message = Message(f"Hello {queue_name}".encode())

            args = ()
            kwargs = {}

            if params_combination == "both_kwargs":
                kwargs = {"message": message, "routing_key": queue_name}
            elif params_combination == "arg_kwarg":
                args = (message,)
                kwargs = {"routing_key": queue_name}
            elif params_combination == "arg_kwarg_empty_key":
                args = (message,)
                kwargs = {"routing_key": ""} 
            else:
                # params_combination == "both_args"
                args = (message, queue_name)

            # Sending the message
            await exchange.publish(*args, **kwargs)

    async def delete_queue(self) -> None:
        connection = await connect()

        async with connection:
            channel = await connection.channel()
            await channel.queue_delete(self.queue_name)

    async def consume_message(self, connect_method) -> None:
        connection = await connect_method()

        async with connection:
            # Creating channel
            channel = await connection.channel()

            # Declaring queue
            queue = await channel.declare_queue(self.queue_name)

            async with queue.iterator() as queue_iter:
                async for message in queue_iter:
                    async with message.process():
                        if queue.name in message.body.decode():
                            break

    async def consume_with_exception(self, connect_method) -> None:
        connection = await connect_method()

        async def on_message(msg):
            raise RuntimeError("Simulated Exception")

        async with connection:
            # Creating channel
            channel = await connection.channel()

            # Declaring queue
            queue = await channel.declare_queue(self.queue_name)

            await queue.consume(on_message)
            await asyncio.sleep(1)  # Wait to ensure the message is processed

    def assert_span_info(self, rabbitmq_span: "ReadableSpan", sort: str, key: str = "test.queue") -> None:
        assert rabbitmq_span.data["rabbitmq"]["exchange"] == "test.exchange"
        assert rabbitmq_span.data["rabbitmq"]["sort"] == sort
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["key"] == key
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

    @pytest.mark.parametrize(
        "params_combination",
        ["both_args", "both_kwargs", "arg_kwarg"],
    )
    def test_basic_publish(self, params_combination) -> None:
        with tracer.start_as_current_span("test"):
            self.loop.run_until_complete(self.publish_message(params_combination))

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        rabbitmq_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == rabbitmq_span.t

        # Parent relationships
        assert rabbitmq_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rabbitmq_span.ec

        # Span attributes
        key = "" if params_combination == "arg_kwarg_empty_key" else self.queue_name
        self.assert_span_info(rabbitmq_span, "publish", key)

    def test_basic_publish_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.loop.run_until_complete(self.publish_message())

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # Parent relationships
        assert not rabbitmq_span.p

        # Error logging
        assert not rabbitmq_span.ec

        # Span attributes
        self.assert_span_info(rabbitmq_span, "publish")

    @pytest.mark.parametrize(
        "connect_method",
        [connect, connect_robust],
    )
    def test_basic_consume(self, connect_method) -> None:
        with tracer.start_as_current_span("test"):
            self.loop.run_until_complete(self.publish_message())
            self.loop.run_until_complete(self.consume_message(connect_method))

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        rabbitmq_publisher_span = spans[0]
        rabbitmq_consumer_span = spans[1]
        test_span = spans[2]

        # Same traceId
        assert test_span.t == rabbitmq_publisher_span.t
        assert rabbitmq_publisher_span.t == rabbitmq_consumer_span.t

        # Parent relationships
        assert rabbitmq_publisher_span.p == test_span.s
        assert rabbitmq_consumer_span.p == rabbitmq_publisher_span.s

        # Error logging
        assert not rabbitmq_publisher_span.ec
        assert not rabbitmq_consumer_span.ec
        assert not test_span.ec

        # Span attributes
        self.assert_span_info(rabbitmq_publisher_span, "publish")
        self.assert_span_info(rabbitmq_consumer_span, "consume")

    @pytest.mark.parametrize(
        "connect_method",
        [connect, connect_robust],
    )
    def test_consume_with_exception(self, connect_method) -> None:
        with tracer.start_as_current_span("test"):
            self.loop.run_until_complete(self.publish_message())
            self.loop.run_until_complete(self.consume_with_exception(connect_method))

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        rabbitmq_publisher_span = spans[0]
        rabbitmq_consumer_span = spans[1]
        test_span = spans[2]

        # Same traceId
        assert test_span.t == rabbitmq_publisher_span.t
        assert rabbitmq_publisher_span.t == rabbitmq_consumer_span.t

        # Parent relationships
        assert rabbitmq_publisher_span.p == test_span.s
        assert rabbitmq_consumer_span.p == rabbitmq_publisher_span.s

        # Error logging
        assert not rabbitmq_publisher_span.ec
        assert rabbitmq_consumer_span.ec == 1
        assert not test_span.ec

        # Span attributes
        self.assert_span_info(rabbitmq_publisher_span, "publish")
        self.assert_span_info(rabbitmq_consumer_span, "consume")



# (c) Copyright IBM Corp. 2025

import yaml

from instana.log import logger


class ConfigReader:
    def __init__(self, file_path: str) -> None:
        self.file_path = file_path
        self.data = {}
        if file_path:
            self.load_file()
        else:
            logger.warning("ConfigReader: No configuration file specified")

    def load_file(self) -> None:
        """Loads and parses the YAML file"""
        try:
            with open(self.file_path, "r") as file:
                self.data = yaml.safe_load(file)
        except FileNotFoundError:
            logger.error(
                f"ConfigReader: Configuration file has not found: {self.file_path}"
            )
        except yaml.YAMLError as e:
            logger.error(f"ConfigReader: Error parsing YAML file: {e}")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import logging
import sys
from collections.abc import Mapping
from typing import Any, Callable, Dict, Tuple

import wrapt

from instana.log import logger
from instana.singletons import agent
from instana.util.runtime import get_runtime_env_info
from instana.util.traceutils import get_tracer_tuple, tracing_is_off


@wrapt.patch_function_wrapper("logging", "Logger._log")
def log_with_instana(
    wrapped: Callable[..., None],
    instance: logging.Logger,
    argv: Tuple[int, str, Tuple[Any, ...]],
    kwargs: Dict[str, Any],
) -> Callable[..., None]:
    # argv[0] = level
    # argv[1] = message
    # argv[2] = args for message

    # We take into consideration if `stacklevel` is already present in `kwargs`.
    # This prevents the error `_log() got multiple values for keyword argument 'stacklevel'`
    stacklevel_in = kwargs.pop(
        "stacklevel", 1 if get_runtime_env_info()[0] not in ["ppc64le", "s390x"] else 2
    )
    stacklevel = stacklevel_in + 1

    try:
        # Only needed if we're tracing and serious log and logging spans are not disabled
        if (
            tracing_is_off()
            or argv[0] < logging.WARN
            or agent.options.is_span_disabled(category="logging")
        ):
            return wrapped(*argv, **kwargs, stacklevel=stacklevel)

        tracer, parent_span, _ = get_tracer_tuple()

        msg = str(argv[1])
        args = argv[2]
        if args and len(args) == 1 and isinstance(args[0], Mapping) and args[0]:
            args = args[0]

        # get the formatted log message
        msg = msg % args

        # get additional information if an exception is being handled
        parameters = None
        (t, v, tb) = sys.exc_info()
        if t is not None and v is not None:
            parameters = "{} {}".format(t, v)

        parent_context = parent_span.get_span_context() if parent_span else None

        # create logging span
        with tracer.start_as_current_span("log", span_context=parent_context) as span:
            event_attributes = {"message": msg}
            if parameters is not None:
                event_attributes.update({"parameters": parameters})
            span.add_event(name="log_with_instana", attributes=event_attributes)
            # extra tags for an error
            if argv[0] >= logging.ERROR:
                span.mark_as_errored()

    except Exception:
        logger.debug("log_with_instana:", exc_info=True)

    return wrapped(*argv, **kwargs, stacklevel=stacklevel)


logger.debug("Instrumenting logging")



# (c) Copyright IBM Corp. 2025

import pytest
import httpx
from typing import Generator
import asyncio

from instana.singletons import agent, tracer
from instana.util.ids import hex_id
import tests.apps.flask_app
from tests.helpers import testenv


@pytest.mark.parametrize("request_mode", ["sync", "async"])
class TestHttpxClients:
    @classmethod
    def setup_class(cls) -> None:
        cls.client = httpx.Client()
        cls.host = "127.0.0.1"
        cls.recorder = tracer.span_processor

    def teardown_class(cls) -> None:
        cls.client.close()

    @pytest.fixture(autouse=True)
    def _resource(self, request_mode) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder.clear_spans()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        yield
        # teardown
        if self.loop.is_running():
            self.loop.close()
        # Ensure that allow_exit_as_root has the default value
        agent.options.allow_exit_as_root = False

    async def get_async_response(self, path, request_method, headers) -> httpx.Response:
        """Asynchronous request function"""
        async with httpx.AsyncClient() as client:
            if request_method == "GET":
                response = await client.get(
                    testenv["flask_server"] + path, headers=headers
                )
            elif request_method == "POST":
                response = await client.post(
                    testenv["flask_server"] + path, headers=headers
                )
            return response

    # Synchronous request function
    def get_sync_response(self, path, request_method, headers) -> httpx.Response:
        """Synchronous request function"""
        if request_method == "GET":
            response = self.client.get(testenv["flask_server"] + path, headers=headers)
        elif request_method == "POST":
            response = self.client.post(testenv["flask_server"] + path, headers=headers)
        return response

    def execute_request(
        self, request_mode, path, request_method="GET", headers=None
    ) -> httpx.Response:
        if request_mode == "async":
            res = self.loop.run_until_complete(
                self.get_async_response(path, request_method, headers)
            )
        elif request_mode == "sync":
            res = self.get_sync_response(path, request_method, headers)
        return res

    def test_get_request(self, request_mode) -> None:
        path = "/"
        with tracer.start_as_current_span("test"):
            res = self.execute_request(request_mode, path)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        httpx_span = spans[1]
        test_span = spans[2]

        assert res
        assert res.status_code == 200

        assert "X-INSTANA-T" in res.headers
        assert int(res.headers["X-INSTANA-T"], 16)
        assert res.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in res.headers
        assert int(res.headers["X-INSTANA-S"], 16)
        assert res.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in res.headers
        assert res.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in res.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert res.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == httpx_span.t
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert httpx_span.p == test_span.s
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not test_span.ec
        assert not httpx_span.ec
        assert not wsgi_span.ec

        # span names
        assert wsgi_span.n == "wsgi"
        assert test_span.data["sdk"]["name"] == "test"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 200
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == "/"
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert httpx_span.data["http"]["method"] == "GET"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

    def test_get_request_as_root_exit_span(self, request_mode) -> None:
        path = "/"
        agent.options.allow_exit_as_root = True
        res = self.execute_request(request_mode, path)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        wsgi_span = spans[0]
        httpx_span = spans[1]

        assert res
        assert res.status_code == 200

        assert "X-INSTANA-T" in res.headers
        assert int(res.headers["X-INSTANA-T"], 16)
        assert res.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in res.headers
        assert int(res.headers["X-INSTANA-S"], 16)
        assert res.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in res.headers
        assert res.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in res.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert res.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert not httpx_span.p
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not httpx_span.ec
        assert not wsgi_span.ec

        # span names
        assert wsgi_span.n == "wsgi"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 200
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == path
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + path
        assert httpx_span.data["http"]["method"] == "GET"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

    def test_get_request_with_query(self, request_mode) -> None:
        path = "/"
        with tracer.start_as_current_span("test"):
            res = self.execute_request(
                request_mode, path + "?user=instana&pass=itsasecret"
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        httpx_span = spans[1]
        test_span = spans[2]

        assert res
        assert res.status_code == 200

        # Same traceId
        assert test_span.t == httpx_span.t
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert httpx_span.p == test_span.s
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not test_span.ec
        assert not httpx_span.ec
        assert not wsgi_span.ec

        # span names
        assert wsgi_span.n == "wsgi"
        assert test_span.data["sdk"]["name"] == "test"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 200
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == path
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + path
        assert httpx_span.data["http"]["params"] == "user=instana&pass=<redacted>"
        assert httpx_span.data["http"]["method"] == "GET"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

    def test_post_request(self, request_mode) -> None:
        path = "/notfound"
        with tracer.start_as_current_span("test"):
            res = self.execute_request(request_mode, path, request_method="POST")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        httpx_span = spans[1]
        test_span = spans[2]

        assert res
        assert res.status_code == 404

        # Same traceId
        assert test_span.t == httpx_span.t
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert httpx_span.p == test_span.s
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not test_span.ec
        assert not httpx_span.ec
        assert not wsgi_span.ec

        # span names
        assert wsgi_span.n == "wsgi"
        assert test_span.data["sdk"]["name"] == "test"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 404
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == path
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + path
        assert httpx_span.data["http"]["method"] == "POST"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

    def test_5xx_request(self, request_mode) -> None:
        path = "/500"
        with tracer.start_as_current_span("test"):
            res = self.execute_request(request_mode, path)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        httpx_span = spans[1]
        test_span = spans[2]

        assert res
        assert res.status_code == 500

        assert "X-INSTANA-T" in res.headers
        assert int(res.headers["X-INSTANA-T"], 16)
        assert res.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in res.headers
        assert int(res.headers["X-INSTANA-S"], 16)
        assert res.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in res.headers
        assert res.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in res.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert res.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == httpx_span.t
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert httpx_span.p == test_span.s
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not test_span.ec
        assert httpx_span.ec == 1
        assert wsgi_span.ec == 1

        # span names
        assert wsgi_span.n == "wsgi"
        assert test_span.data["sdk"]["name"] == "test"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 500
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == path
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + path
        assert httpx_span.data["http"]["method"] == "GET"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

    def test_response_header_capture(self, request_mode) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]
        path = "/response_headers"

        with tracer.start_as_current_span("test"):
            res = self.execute_request(request_mode, path)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        httpx_span = spans[1]
        test_span = spans[2]

        assert res
        assert res.status_code == 200

        # Same traceId
        assert test_span.t == httpx_span.t
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert httpx_span.p == test_span.s
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not test_span.ec
        assert not httpx_span.ec
        assert not wsgi_span.ec

        # span names
        assert wsgi_span.n == "wsgi"
        assert test_span.data["sdk"]["name"] == "test"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 200
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == path
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + path
        assert httpx_span.data["http"]["method"] == "GET"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

        assert "X-Capture-This" in httpx_span.data["http"]["header"]
        assert httpx_span.data["http"]["header"]["X-Capture-This"] == "Ok"
        assert "X-Capture-That" in httpx_span.data["http"]["header"]
        assert httpx_span.data["http"]["header"]["X-Capture-That"] == "Ok too"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture(self, request_mode) -> None:
        path = "/"
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        request_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }
        with tracer.start_as_current_span("test"):
            res = self.execute_request(request_mode, path, headers=request_headers)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        httpx_span = spans[1]
        test_span = spans[2]

        assert res
        assert res.status_code == 200

        # Same traceId
        assert test_span.t == httpx_span.t
        assert httpx_span.t == wsgi_span.t

        # Parent relationships
        assert httpx_span.p == test_span.s
        assert wsgi_span.p == httpx_span.s

        # Error logging
        assert not test_span.ec
        assert not httpx_span.ec
        assert not wsgi_span.ec

        # span names
        assert wsgi_span.n == "wsgi"
        assert test_span.data["sdk"]["name"] == "test"
        assert httpx_span.n == "http"

        # httpx
        assert httpx_span.data["http"]["status"] == 200
        assert httpx_span.data["http"]["host"] == self.host
        assert httpx_span.data["http"]["path"] == "/"
        assert httpx_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert httpx_span.data["http"]["method"] == "GET"
        assert httpx_span.stack
        assert isinstance(httpx_span.stack, list)
        assert len(httpx_span.stack) > 1

        assert "X-Capture-This-Too" in httpx_span.data["http"]["header"]
        assert httpx_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in httpx_span.data["http"]["header"]
        assert httpx_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import importlib.metadata
import json
from collections import defaultdict
from typing import Any, DefaultDict
from urllib import parse

from instana.log import logger


def nested_dictionary() -> DefaultDict[str, Any]:
    return defaultdict(DictionaryOfStan)


# Simple implementation of a nested dictionary.
DictionaryOfStan: DefaultDict[str, Any] = nested_dictionary


# Assisted by watsonx Code Assistant
def to_json(obj: Any) -> bytes:
    """
    Convert the given object to a JSON binary string.

    This function is primarily used to serialize objects from `json_span.py`
    until a switch to nested dictionaries (or a better solution) is made.

    :param obj: The object to serialize to JSON.
    :return: The JSON string encoded as bytes.
    """
    try:

        def extractor(o: Any) -> dict:
            """
            Extract dictionary-like attributes from an object.

            :param o: The object to extract attributes from.
            :return: A dictionary containing the object's attributes.
            """
            if not hasattr(o, "__dict__"):
                logger.debug(f"Couldn't serialize non dict type: {type(o)}")
                return {}
            else:
                return {k.lower(): v for k, v in o.__dict__.items() if v is not None}

        return json.dumps(
            obj, default=extractor, sort_keys=False, separators=(",", ":")
        ).encode()
    except Exception:
        logger.debug("to_json non-fatal encoding issue: ", exc_info=True)


# Assisted by watsonx Code Assistant
def to_pretty_json(obj: Any) -> str:
    """
    Convert an object to a pretty-printed JSON string.

    This function is primarily used for logging and debugging purposes.

    :param obj: the object to serialize to json
    :return:  json string
    """
    try:

        def extractor(o):
            if not hasattr(o, "__dict__"):
                logger.debug("Couldn't serialize non dict type: %s", type(o))
                return {}
            else:
                return {k.lower(): v for k, v in o.__dict__.items() if v is not None}

        return json.dumps(
            obj, default=extractor, sort_keys=True, indent=4, separators=(",", ":")
        )
    except Exception:
        logger.debug("to_pretty_json non-fatal encoding issue: ", exc_info=True)


# Assisted by watsonx Code Assistant
def package_version() -> str:
    """
    Determine the version of the 'instana' package.

    This function uses the `importlib.metadata` module to fetch the version of
    the 'instana' package.
    If the package is not found, it returns 'unknown'.

    :return: A string representing the version of the 'instana' package.
    """
    try:
        version = importlib.metadata.version("instana")
    except importlib.metadata.PackageNotFoundError:
        logger.debug("Not able to identify the Instana package version.")
        version = "unknown"

    return version


# Assisted by watsonx Code Assistant
def get_default_gateway() -> str:
    """
    Attempts to read /proc/self/net/route to determine the default gateway in use.

    This function reads the /proc/self/net/route file, which contains network
    routing information for the current process.
    It specifically looks for the line where the Destination is 00000000,
    indicating the default route.
    The Gateway IP is encoded backwards in hex, which this function decodes and
    converts to a standard IP address format.

    :return: String -   the ip address of the default gateway or None if not
                        found/possible/non-existant
    """
    try:
        hip = None
        # The first line is the header line
        # We look for the line where the Destination is 00000000 - that is the default route
        # The Gateway IP is encoded backwards in hex.
        with open("/proc/self/net/route") as routes:
            for line in routes:
                parts = line.split("\t")
                if parts[1] == "00000000":
                    hip = parts[2]

        if hip is not None and len(hip) == 8:
            # Reverse order, convert hex to int
            return f"{int(hip[6:8], 16)}.{int(hip[4:6], 16)}.{int(hip[2:4], 16)}.{int(hip[0:2], 16)}"

    except Exception:
        logger.warning("get_default_gateway: ", exc_info=True)


# Assisted by watsonx Code Assistant
def validate_url(url: str) -> bool:
    """
    Validate if the provided <url> is a valid URL.

    This function checks if the given string is a valid URL by attempting to
    parse it using the `urlparse` function from the `urllib.parse` module.
    A URL is considered valid if it has both a scheme (like 'http' or 'https')
    and a network location (netloc).

    Examples:
    - "http://localhost:5000" - valid
    - "http://localhost:5000/path" - valid
    - "sandwich" - invalid

    @param url: A string representing the URL to validate.
    @return: A boolean value. Returns `True` if the URL is valid, otherwise `False`.
    """
    try:
        result = parse.urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception:
        pass

    return False
